/* This is an example of the optimizer getting the number of rows wrong because the data is 
skewed.  It is fixed by using histograms */

set linesize 1000
set pagesize 1000
set echo on
set termout on

spool test2.log

drop table TEST2;

create table TEST2 (a number);

declare
  i number;
begin
  i:=1;
  loop
    insert into TEST2 values (1);
    i:=i+1;
    exit when (i > 1000000);
    if (mod(i,10000)=0) then
      commit;
    end if;
  end loop;
  commit;
  insert into TEST2 values (2);
  commit;
end;
/

create index test2index on TEST2 (a); 

/* Regular gather stats compute without histograms */

execute DBMS_STATS.GATHER_TABLE_STATS (NULL,'TEST2',cascade=>TRUE,estimate_percent=>NULL,method_opt=>'FOR ALL COLUMNS SIZE 1');

drop table plan_table;

@utlxplan

explain plan into plan_table for 
select * from TEST2 where a=2;

set echo off
@utlxpls.sql
set echo on

/* Show how long it takes with the wrong plan */

set timing on 

select * from TEST2 where a=2;

set timing off

/* Show the low and high values and number of distinct values that optimizer has available. 
Show that there are no histograms = one bucket.  */

select UTL_RAW.CAST_TO_NUMBER(low_value) Low,UTL_RAW.CAST_TO_NUMBER(high_value) High,
num_distinct,num_buckets
from USER_TAB_COLUMNS 
where table_name='TEST2' and 
column_name='A';

/* Show the number of rows in the table statistic */

select num_rows
from USER_TABLES 
where table_name='TEST2';

/* Drop and recreate table to clear out previous execution plan - otherwise it says it is
using the range scan but doesn't! */

drop table TEST2;

create table TEST2 (a number);

declare
  i number;
begin
  i:=1;
  loop
    insert into TEST2 values (1);
    i:=i+1;
    exit when (i > 1000000);
    if (mod(i,10000)=0) then
      commit;
    end if;
  end loop;
  commit;
  insert into TEST2 values (2);
  commit;
end;
/

create index test2index on TEST2 (a); 

/* Gather stats with maximum number of histograms */

execute DBMS_STATS.GATHER_TABLE_STATS (NULL,'TEST2',cascade=>TRUE,estimate_percent=>NULL,method_opt=>'FOR ALL COLUMNS SIZE 254');

drop table plan_table;

@utlxplan

explain plan into plan_table for 
select * from TEST2 where a=2;

set echo off
@utlxpls.sql
set echo on

/* Show how long it takes with the right plan */

set timing on 

select * from TEST2 where a=2;

set timing off

/* Show the low and high values and number of distinct values that optimizer has available. 
Show that there are histograms = 2 buckets.  */

select UTL_RAW.CAST_TO_NUMBER(low_value) Low,UTL_RAW.CAST_TO_NUMBER(high_value) High,
num_distinct,num_buckets
from USER_TAB_COLUMNS 
where table_name='TEST2' and 
column_name='A';

/* Show distribution of data in the table */

select a,count(*) from TEST2 group by a;

drop table TEST2;

spool off
